



pip check || exit 0
IF %ERRORLEVEL% NEQ 0 exit /B 1
datashader --help
IF %ERRORLEVEL% NEQ 0 exit /B 1
datashader copy-examples --path=. --force
IF %ERRORLEVEL% NEQ 0 exit /B 1
pytest --nbsmoke-run -k ".ipynb" getting_started/2_Pipeline.ipynb
IF %ERRORLEVEL% NEQ 0 exit /B 1
exit /B 0
